/*
 * Copyright (c) 2012 Memorial-Sloan Kettering Cancer Center.
 *
 * This file is part of BP_Prior.
 *
 * BP_Prior is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BP_Prior is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with BP_Prior.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mskcc.cbio.piclub;

import org.apache.commons.cli.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.mskcc.cbio.piclub.model.ModelNode;
import org.mskcc.cbio.piclub.util.ModelNodeUtil;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.*;

/* BP network inference specific post-cleaning script */
public class CleanDistanceNetwork {
    private final static Log log = LogFactory.getLog(CleanDistanceNetwork.class);
    private final static String className = CleanDistanceNetwork.class.getSimpleName();
    private final static String helpText = className + " node_list.tsv old_network.tsv new_network.tsv";

    public static void main(String[] args) throws FileNotFoundException {
        final CommandLineParser commandLineParser = new GnuParser();
        final Options gnuOptions = new Options();

        try {
            CommandLine commandLine = commandLineParser.parse(gnuOptions, args);
            String nodeFile;
            String networkFile;
            String newNetworkFile;

            final String[] others = commandLine.getArgs();
            if(others.length >= 3) {
                nodeFile = others[0];
                networkFile = others[1];
                newNetworkFile = others[2];
            } else {
                throw new ParseException("too few arguments.");
            }

            List<ModelNode> modelNodes = ModelNodeUtil.parseGeneFile(new FileInputStream(nodeFile));
            Map<String, ModelNode> nameToModelNode = new HashMap<String, ModelNode>();
            Map<String, String[]> lineMap = new HashMap<String, String[]>();

            for (ModelNode node : modelNodes)
                nameToModelNode.put(node.getName(), node);

            Scanner networkScanner = new Scanner(new FileInputStream(networkFile));
            while(networkScanner.hasNextLine()) {
                String line = networkScanner.nextLine();
                if(line.isEmpty())
                    continue;

                String tokens[] = line.split("\t");
                String srcName = tokens[0];
                String distance = tokens[1];
                String targetName = tokens[2];
                String intTypes = tokens[5];

                ModelNode targetNode = nameToModelNode.get(targetName);

                boolean isTargetActNode = isActivationNode(targetNode);

                // Rule 1: If it is a self edge, remove it
                if(srcName.equals(targetName)) {
                    log.info("Removing self-edge: " + line);
                    continue;
                }

                // Rule 2: If the target is a concentration node and if we don't have complex formation remove it
                if(targetNode.getType().equals(ModelNode.Type.CONCENTRATION) && !intTypes.contains("ComplexAssembly")) {
                    log.info("Removing concentration node targeted edge (no complex formation): " + line);
                    continue;
                }

                // Rule 3: If the target is an activation node, remove it
                if(isTargetActNode) {
                    log.info("Removing edge (activation node as target): " + line);
                    continue;
                }

                String key = srcName + distance + targetName;
                String[] strings = lineMap.get(key);
                if(strings != null) {
                    for(int i = 3; i <= 5; i++)
                        tokens[i] = takeUnion(tokens[i], strings[i]);
                }

                lineMap.put(key, tokens);
            }

            PrintStream ps = new PrintStream(new FileOutputStream(newNetworkFile));
            for (String[] tokens : lineMap.values()) {
                String aLine = "";
                for (String token : tokens)
                    aLine += token + "\t";

                aLine = aLine.substring(0, aLine.length()-1);
                ps.println(aLine);
            }
            ps.close();

            log.info("Done!");

        } catch (Exception e) {
            System.err.println(e.getMessage());
            HelpFormatter helpFormatter = new HelpFormatter();
            helpFormatter.printHelp(helpText, gnuOptions);
            System.exit(-1);
        }
    }

    private static String takeUnion(String a, String b) {
        String sep = ";";
        HashSet<String> set = new HashSet<String>(Arrays.asList(a.split(sep)));
        set.addAll(Arrays.asList(b.split(sep)));

        String finalStr = "";
        for (String s : set) {
            finalStr += s + ";";
        }

        if(finalStr.length() > 0)
            finalStr = finalStr.substring(0, finalStr.length()-1);

        return finalStr;
    }

    private static boolean isActivationNode(ModelNode node) {
        String name = node.getName();

        return name.startsWith("a")
                && Character.isUpperCase(name.charAt(1))
                && node.getType().equals(ModelNode.Type.ACTIVATED);
    }
}
