/*
 * Copyright (c) 2012 Memorial-Sloan Kettering Cancer Center.
 *
 * This file is part of BP_Prior.
 *
 * BP_Prior is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BP_Prior is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with BP_Prior.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mskcc.cbio.piclub;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.biopax.paxtools.io.sif.level3.ActivityNetworkAnalyzer;
import org.biopax.paxtools.io.sif.level3.ChangeType;
import org.biopax.paxtools.io.sif.level3.PEStateChange;
import org.biopax.paxtools.model.Model;
import org.biopax.paxtools.model.level3.EntityFeature;
import org.biopax.paxtools.model.level3.EntityReference;
import org.biopax.paxtools.model.level3.RelationshipXref;
import org.biopax.paxtools.model.level3.SimplePhysicalEntity;
import org.mskcc.cbio.piclub.model.ModelNode;
import org.mskcc.cbio.piclub.model.Phosphorylation;
import org.mskcc.cbio.piclub.util.PathwayCommons2Util;
import org.mskcc.cbio.piclub.util.RelationshipXrefExtractor;

import java.util.*;

public class ModelNodeMapper {
    private static Log log = LogFactory.getLog(ModelNodeMapper.class);
    private boolean strict = false;
    private int siteMismatchTolerance = PathwayCommons2Util.PHOSPHORYLATION_SITE_MISMATCH_TOLERANCE;
    Map<ModelNode, Set<String>> blacklists = null;

    public Map<ModelNode, Set<SimplePhysicalEntity>> map(Model model, Collection<ModelNode> modelNodes) {
        log.info("Mapping model nodes to the pathway entities.");
        Map<ModelNode, Set<SimplePhysicalEntity>> nodeToEntity = new HashMap<ModelNode, Set<SimplePhysicalEntity>>();
        HashSet<SimplePhysicalEntity> simplePhysicalEntities
                = new HashSet<SimplePhysicalEntity>(model.getObjects(SimplePhysicalEntity.class));
        blacklists = new HashMap<ModelNode, Set<String>>();

        RelationshipXrefExtractor xrefExtractor = new RelationshipXrefExtractor();
        ActivityNetworkAnalyzer activityNetworkAnalyzer = new ActivityNetworkAnalyzer();
        activityNetworkAnalyzer.analyzeStates(model);

        Map<String, Set<ModelNode>> symbolToNodes = new HashMap<String, Set<ModelNode>>();
        for(ModelNode modelNode : modelNodes) {
            String tmpSymbol = modelNode.getGeneSymbol();
            Set<ModelNode> tmpNodes = symbolToNodes.get(tmpSymbol);
            if(tmpNodes == null) {
                tmpNodes = new HashSet<ModelNode>();
                symbolToNodes.put(tmpSymbol, tmpNodes);
            }

            tmpNodes.add(modelNode);
        }

        HashSet<String> symbols = new HashSet<String>(symbolToNodes.keySet());
        for(String symbol: symbols) {
            Set<ModelNode> symbolNodes = symbolToNodes.get(symbol);
            Set<ModelNode> tmpNodes = new HashSet<ModelNode>(symbolNodes);
            for(ModelNode modelNode : tmpNodes) {
                RelationshipXref rxref = (RelationshipXref) model.getByID(PathwayCommons2Util.convertToURI(symbol));
                if(rxref == null)
                    continue;

                Phosphorylation phosphorylation = modelNode.getPhosphorylation();
                Set<SimplePhysicalEntity> entities = new HashSet<SimplePhysicalEntity>();

                log.info("[" + modelNode.getName() + " (" + symbol + ")"
                        + (!phosphorylation.isNone() ? "_p@" + phosphorylation : "")
                        + "|" + modelNode.getType()
                        + "]"
                );

                for (EntityReference er : xrefExtractor.getERs(rxref)) {
                    Set<PEStateChange> speStates = activityNetworkAnalyzer.getAllStates(er);
                    if(speStates == null)
                        continue;

                    for (PEStateChange speState : speStates) {
                        Map<EntityFeature, ChangeType> deltaFeatures = speState.getDeltaFeatures();
                        Set<EntityFeature> disappearingFeatures = new HashSet<EntityFeature>();
                        Set<EntityFeature> appearingFeatures = new HashSet<EntityFeature>();

                        for (EntityFeature entityFeature : deltaFeatures.keySet()) {
                            ChangeType changeType = deltaFeatures.get(entityFeature);
                            switch (changeType) {
                                case EXIST_TO_NOT_EXIST:
                                case EXIST_TO_UNKNOWN:
                                    disappearingFeatures.add(entityFeature);
                                    break;
                                case NOT_EXIST_TO_EXIST:
                                case UNKNOWN_TO_EXIST:
                                    appearingFeatures.add(entityFeature);
                                    break;
                            }
                        }
                        SimplePhysicalEntity spe = null;

                        if(PathwayCommons2Util.hasEntityFeatures(modelNode, disappearingFeatures, strict))
                            spe = speState.getLeft();
                        else if(PathwayCommons2Util.hasEntityFeatures(modelNode, appearingFeatures, strict))
                            spe = speState.getRight();

                        if(spe != null && !entities.contains(spe)) {
                            entities.add(spe);
                            log.info("\t\t*\t" + PathwayCommons2Util.entityToString(spe) + "\t" + spe.getRDFId());
                        }
                    }
                }

                Set<String> blacklist = new HashSet<String>();
                for(SimplePhysicalEntity spe: simplePhysicalEntities) {
                    Set<RelationshipXref> xrefs = xrefExtractor.getXrefs(model, spe);
                    if(!xrefs.contains(rxref) || entities.contains(spe))
                        continue;

                    if(PathwayCommons2Util.hasEquivalentStates(modelNode, spe, strict)) {
                        entities.add(spe);
                        log.info("\t\t\t" + PathwayCommons2Util.entityToString(spe) + "\t" + spe.getRDFId());
                    } else {
                        blacklist.add(spe.getRDFId());
                    }
                }
                blacklists.put(modelNode, blacklist);

                if(entities.isEmpty()) {
                    log.warn("Could not map " + modelNode + " to a PC2 entity.");
                    symbolNodes.remove(modelNode);
                    if(symbolNodes.isEmpty())
                        symbolToNodes.remove(symbol);
                } else {
                    nodeToEntity.put(modelNode, entities);
                }
            }
        }

        return nodeToEntity;
    }

    public boolean isStrict() {
        return strict;
    }

    public void setStrict(boolean strict) {
        this.strict = strict;
    }

    public int getSiteMismatchTolerance() {
        return siteMismatchTolerance;
    }

    public void setSiteMismatchTolerance(int siteMismatchTolerance) {
        this.siteMismatchTolerance = siteMismatchTolerance;
        PathwayCommons2Util.PHOSPHORYLATION_SITE_MISMATCH_TOLERANCE = siteMismatchTolerance;
    }

    public Map<ModelNode, Set<String>> getBlacklists() {
        return blacklists;
    }
}
