/*
 * Copyright (c) 2012 Memorial-Sloan Kettering Cancer Center.
 *
 * This file is part of BP_Prior.
 *
 * BP_Prior is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BP_Prior is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with BP_Prior.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mskcc.cbio.piclub;

import org.apache.commons.cli.*;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.biopax.paxtools.controller.Cloner;
import org.biopax.paxtools.controller.Completer;
import org.biopax.paxtools.io.SimpleIOHandler;
import org.biopax.paxtools.model.BioPAXElement;
import org.biopax.paxtools.model.Model;
import org.biopax.paxtools.model.level3.*;
import org.biopax.paxtools.query.QueryExecuter;
import org.biopax.paxtools.query.algorithm.Direction;
import org.mskcc.cbio.piclub.model.Distance;
import org.mskcc.cbio.piclub.model.ModelNode;
import org.mskcc.cbio.piclub.util.ModelNodeUtil;
import org.mskcc.cbio.piclub.util.DistanceUtil;
import org.mskcc.cbio.piclub.util.PathwayCommons2Util;

import java.io.*;
import java.util.*;

public class PriorMain {
    private final static Log log = LogFactory.getLog(PriorMain.class);
    private final static String className = PriorMain.class.getSimpleName();

    public static void main(String args[]) throws IOException {
        String OUTPUT_FILE = "prior_network.tsv";
        String NODE_FILE = "";
        String SAVE_PATH_FOLDER = "";
        boolean STRICT_MODE = false;
        boolean BLACKLIST_ENABLED = false;
        String MODEL_FILE = "";
        Integer GRAPH_LIMIT = 2;
        boolean SAVE_PATHS = false;
        final String helpText = className + " node_list.tsv biopax_model.owl";

        final CommandLineParser commandLineParser = new GnuParser();
        final Options gnuOptions = new Options();
        gnuOptions.addOption("o", "output-file", true, "the output file (default ./prior_network.tsv)")
                .addOption("l", "graph-limit", true, "distance limit to be used in graph queries (default: " + GRAPH_LIMIT + ")")
                .addOption("t", "site-tolerance", true, "phosphorylation site mismatch tolerance (default: " + PathwayCommons2Util.PHOSPHORYLATION_SITE_MISMATCH_TOLERANCE +")")
                .addOption("s", "strict", false, "turns on strict node to entity mapping, requires exact match of phosphorylation residues (default: false)")
                .addOption("p", "save-paths", true, "if given, the shortest paths between nodes will be saved under the provided directory as BioPAX files (good for debugging; default: do not save)")
                .addOption("b", "blacklist", false, "when enabled, it excludes the paths that are passing through non-matching entities (default: false)")
        ;
        try {
            CommandLine commandLine = commandLineParser.parse(gnuOptions, args);
            if(commandLine.hasOption("o")) {
                OUTPUT_FILE = commandLine.getOptionValue("o");
            }

            if(commandLine.hasOption("l")) {
                GRAPH_LIMIT = Integer.parseInt(commandLine.getOptionValue("l"));
            }

            if(commandLine.hasOption("t")) {
                PathwayCommons2Util.PHOSPHORYLATION_SITE_MISMATCH_TOLERANCE
                        = Integer.parseInt(commandLine.getOptionValue("t"));
            }

            if(commandLine.hasOption("p")) {
                SAVE_PATH_FOLDER = commandLine.getOptionValue("p");
                File savePathFolder = new File(SAVE_PATH_FOLDER);
                if(!savePathFolder.exists() && !savePathFolder.mkdirs()) {
                    System.err.println("Could not create the folder: " + SAVE_PATH_FOLDER);
                    System.exit(-1);
                }
            }

            STRICT_MODE |= commandLine.hasOption("s");
            BLACKLIST_ENABLED |= commandLine.hasOption("b");
            SAVE_PATHS |= commandLine.hasOption("p");

            final String[] others = commandLine.getArgs();
            if(others.length >= 2) {
                NODE_FILE = others[0];
                MODEL_FILE = others[1];
            } else {
                HelpFormatter helpFormatter = new HelpFormatter();
                helpFormatter.printHelp(helpText, gnuOptions);
                System.exit(-1);
            }

        } catch (ParseException e) {
            System.err.println(e.getMessage());
            HelpFormatter helpFormatter = new HelpFormatter();
            helpFormatter.printHelp(helpText, gnuOptions);
            System.exit(-1);
        }

        log.info("Reading gene list file: " + NODE_FILE);
        List<ModelNode> modelNodes = null;
        try {
            modelNodes = ModelNodeUtil.parseGeneFile(new FileInputStream(NODE_FILE));
        } catch (Exception e) {
            log.error(e.getMessage());
            e.printStackTrace();
            System.exit(-1);
        }
        Set<String> idList = new HashSet<String>(PathwayCommons2Util.convertToURI(modelNodes));
        log.info("Read " + idList.size() + " genes and " + modelNodes.size() + " nodes.");

        log.info("Reading the BioPAX model file: " + MODEL_FILE);
        SimpleIOHandler io = new SimpleIOHandler();
        Model rawModel = io.convertFromOWL(new FileInputStream(MODEL_FILE));
        log.info("Done reading the model file.");


        Map<String, Set<ModelNode>> symbolToNodes = new HashMap<String, Set<ModelNode>>();
        for(ModelNode modelNode : modelNodes) {
            String tmpSymbol = modelNode.getGeneSymbol();
            Set<ModelNode> tmpNodes = symbolToNodes.get(tmpSymbol);
            if(tmpNodes == null) {
                tmpNodes = new HashSet<ModelNode>();
                symbolToNodes.put(tmpSymbol, tmpNodes);
            }

            tmpNodes.add(modelNode);
        }

        Set<String> symbols = new HashSet<String>(ModelNodeUtil.extractGeneSymbols(modelNodes));
        log.info("Check if the model has all the genes?");
        for(String symbol: symbols) {
            BioPAXElement bpe = rawModel.getByID(PathwayCommons2Util.convertToURI(symbol));
            if(bpe == null) {
                symbolToNodes.remove(symbol);
                log.warn("Could not find " + symbol + " on PC2.");
            }
        }
        log.info("Found " + symbolToNodes.size() + "/" + symbols.size() + " genes on PC2.");

        HashSet<BioPAXElement> allXrefs = new HashSet<BioPAXElement>();
        for (String symbol : symbols) {
            RelationshipXref xref = (RelationshipXref) rawModel.getByID(PathwayCommons2Util.convertToURI(symbol));
            allXrefs.add(xref);
        }

        Set<BioPAXElement> result = QueryExecuter.runPathsBetween(allXrefs, rawModel, GRAPH_LIMIT);
        Set<BioPAXElement> partialResults
                = QueryExecuter.runNeighborhood(allXrefs, rawModel, GRAPH_LIMIT, Direction.BOTHSTREAM);
        if(result == null || partialResults == null) {
            System.err.println("Could not connect gene nodes in the BioPAX model. Quiting...");
            System.exit(-1);
        }
        Set<BioPAXElement> allResults = new HashSet<BioPAXElement>();
        allResults.addAll(result);
        allResults.addAll(partialResults);

        Completer c = new Completer(io.getEditorMap());
        allResults = c.complete(allResults, null);
        Cloner cln = new Cloner(io.getEditorMap(), io.getLevel().getDefaultFactory());
        Model model = cln.clone(null, allResults);

        log.info("Check if the model has all genes?");
        symbols = new HashSet<String>(symbolToNodes.keySet());
        for(String symbol: symbols) {
            BioPAXElement bpe = model.getByID(PathwayCommons2Util.convertToURI(symbol));
            if(bpe == null) {
                symbolToNodes.remove(symbol);
                log.warn("Could not find " + symbol + " in the model.");
            }
        }
        log.info("Found " + symbolToNodes.size() + "/" + idList.size() + " genes in the pathsBetween model.");

        log.info("Mapping model nodes to the pathway entities.");
        ModelNodeMapper modelNodeMapper = new ModelNodeMapper();
        Map<ModelNode, Set<SimplePhysicalEntity>> nodeToEntity = modelNodeMapper.map(model, modelNodes);
        modelNodeMapper.setSiteMismatchTolerance(PathwayCommons2Util.PHOSPHORYLATION_SITE_MISMATCH_TOLERANCE);
        modelNodeMapper.setStrict(STRICT_MODE);
        Map<ModelNode, Set<String>> blacklists
                = BLACKLIST_ENABLED
                    ? modelNodeMapper.getBlacklists()
                    : new HashMap<ModelNode, Set<String>>();

        log.info("Getting pairwise distances from the PC2 model...");
        ArrayList<ModelNode> nodeList = new ArrayList<ModelNode>(nodeToEntity.keySet());
        BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(OUTPUT_FILE));
        PrintStream ps = new PrintStream(bos);

        for(ModelNode nodeA: nodeList) {
            for(ModelNode nodeB: nodeList) {
                if(nodeA.equals(nodeB))
                    continue;

                // Check if two nodes have a shared entity mapping. If so the distance
                // would be misleading, so we will skip these types of pair comparisons.
                if(nodeA.getGeneSymbol().equals(nodeB.getGeneSymbol())) {
                    Set<SimplePhysicalEntity> tmpEntities = new HashSet<SimplePhysicalEntity>(nodeToEntity.get(nodeA));
                    tmpEntities.retainAll(nodeToEntity.get(nodeB));

                    if(!tmpEntities.isEmpty()) {
                        log.trace("Skipping " + nodeA.getName() + " vs " + nodeB.getName());
                        continue;
                    }
                }

                Distance distance = PathwayCommons2Util.findMinBioPAXDistance(
                        nodeA,
                        nodeB,
                        nodeToEntity,
                        model,
                        GRAPH_LIMIT,
                        blacklists);

                if(distance.getDistance() > 0 ) {
                    log.info(nodeA.getName() + "\t"
                            + distance.getDistance() + "\t"
                            + nodeB.getName());

                    ps.println(nodeA.getName() + "\t"
                            + distance.getDistance() + "\t"
                            + nodeB.getName() + "\t"
                            + DistanceUtil.extractResources(distance) + "\t"
                            + DistanceUtil.extractPublications(distance) + "\t"
                            + DistanceUtil.extractTypes(distance));

                    if(SAVE_PATHS) {
                        String pathFileName = SAVE_PATH_FOLDER + "/" + nodeA.getName() + "." + nodeB.getName() + ".owl";
                        FileOutputStream tmpFos = new FileOutputStream(pathFileName);
                        io.convertToOWL(distance.getModel(), tmpFos);
                        tmpFos.close();
                    }
                }
            }
        }

        ps.close();
        bos.close();
        
        log.info("Done! Please see " + OUTPUT_FILE + " for results");
    }
}
