/*
 * Copyright (c) 2012 Memorial-Sloan Kettering Cancer Center.
 *
 * This file is part of BP_Prior.
 *
 * BP_Prior is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BP_Prior is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with BP_Prior.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mskcc.cbio.piclub.util;

import org.biopax.paxtools.model.BioPAXElement;
import org.biopax.paxtools.model.BioPAXLevel;
import org.biopax.paxtools.model.Model;
import org.biopax.paxtools.model.level3.*;
import org.biopax.paxtools.query.QueryExecuter;
import org.biopax.paxtools.query.algorithm.LimitType;
import org.biopax.paxtools.query.wrapperL3.UbiqueFilter;
import org.mskcc.cbio.piclub.model.Distance;
import org.mskcc.cbio.piclub.model.ModelNode;
import org.mskcc.cbio.piclub.model.Phosphorylation;

import java.util.*;

public class PathwayCommons2Util {
    public static Integer PHOSPHORYLATION_SITE_MISMATCH_TOLERANCE = 4;
    private static HGNCUtil hgncUtil = HGNCUtil.getInstance();

    public static String convertToURI(String geneSymbol) {
        return "urn:biopax:RelationshipXref:HGNC_HGNC%3A" + hgncUtil.getID(geneSymbol);
    }

    public static List<String> convertToURI(Collection<ModelNode> modelNodes) {
        List<String> uris = new ArrayList<String>();
        for(ModelNode modelNode : modelNodes)
            uris.add(convertToURI(modelNode.getGeneSymbol()));

        return uris;
    }

    public static boolean hasEquivalentStates(ModelNode modelNode, SimplePhysicalEntity spe) {
        return hasEquivalentStates(modelNode, spe, false);
    }

    public static boolean hasEquivalentStates(ModelNode modelNode, SimplePhysicalEntity spe, boolean beStrict) {
        Phosphorylation phosphorylation = modelNode.getPhosphorylation();

        boolean stateMatch = false;
        int matches = 0, mismatches = 0;
        for(EntityFeature ef: spe.getFeature()) {
            if(!(ef instanceof ModificationFeature))
                continue;

            boolean residueMatches = false;

            SequenceModificationVocabulary modificationType = ((ModificationFeature) ef).getModificationType();
            if(modificationType == null) continue;

            Set<String> termSet = modificationType.getTerm();
            if(termSet == null) continue;

            Iterator<String> iterator = termSet.iterator();
            if(!iterator.hasNext()) continue;

            String term = iterator.next().toUpperCase();
            switch(modelNode.getType()) {
                case ACTIVATED:
                    if(term.endsWith("INACTIVE"))
                        return false;
                    else if(term.endsWith("ACTIVE"))
                        stateMatch = residueMatches = true;
                    break;
                case INHIBITED:
                    if(term.endsWith("INACTIVE"))
                        stateMatch = residueMatches = true;
                    else if(term.endsWith("ACTIVE"))
                        return false;
                    break;
                case CONCENTRATION:
                    if(term.endsWith("INACTIVE"))
                        return false;
                default:
                    break;
            }

            if(residueMatches) {
                if(!beStrict)
                    matches++;

                continue;
            }

            switch(phosphorylation.getResidue()) {
                case THREONINE:
                    residueMatches = term.endsWith("phospho-L-threonine".toUpperCase());
                    break;
                case SERINE:
                    residueMatches = term.endsWith("phospho-L-serine".toUpperCase());
                    break;
                case TYROSINE:
                    residueMatches = term.endsWith("phospho-L-tyrosine".toUpperCase());
                    break;
                default:
                    residueMatches = true;
            }

            SequenceLocation sloc = ef.getFeatureLocation();
            Integer loc = 0;
            if(sloc instanceof SequenceSite) {
                SequenceSite ss = (SequenceSite) sloc;
                loc = ss.getSequencePosition();
            }

            residueMatches &= Math.abs(loc - phosphorylation.getSite()) < PHOSPHORYLATION_SITE_MISMATCH_TOLERANCE
                                || loc == 0;

            if(residueMatches)
                matches++;
            else
                mismatches++;
        }

        if(beStrict && matches == 0 && mismatches == 0 && phosphorylation.isNone() && stateMatch)
            matches++;

        return !(beStrict && mismatches > 0)
                    && (matches >= 1
                        || modelNode.getType().equals(ModelNode.Type.CONCENTRATION)
                        && phosphorylation.isNone()
                    );
    }

    public static String entityToString(SimplePhysicalEntity spe) {
        String name = spe.getDisplayName() + " ";

        for(EntityFeature ef: spe.getFeature()) {
            String term = "";
            String loc = "";
            SequenceLocation sl = ef.getFeatureLocation();

            if(sl instanceof SequenceSite) {
                loc = "" + ((SequenceSite) sl).getSequencePosition();
            } else if (sl instanceof SequenceInterval) {
                SequenceInterval si = (SequenceInterval) sl;
                loc =
                   (si.getSequenceIntervalBegin() != null ? si.getSequenceIntervalBegin().getSequencePosition() : 0)
                   + "-"
                   + (si.getSequenceIntervalBegin() != null ? si.getSequenceIntervalBegin().getSequencePosition() : 0);
            }

            if(ef instanceof ModificationFeature) {
                SequenceModificationVocabulary modificationType = ((ModificationFeature) ef).getModificationType();
                if(modificationType == null) {
                    term = "null";
                } else {
                    Set<String> termSet = modificationType.getTerm();
                    if(termSet == null) {
                        term = "null";
                    } else if (!termSet.isEmpty()) {
                        term = termSet.iterator().next();
                    } else {
			term = "null";
		   }
                }
            } else if(ef instanceof FragmentFeature) {
                term = "fragmentation";
            } else if(ef instanceof BindingFeature) {
                term = "binds " +
                        ((BindingFeature) ef).getBindsTo().getFeatureOf().iterator().next().getDisplayName();
            }

            name += "(" + term + loc + ") ";
        }

        return name;
    }

    public static Distance findMinBioPAXDistance(ModelNode nodeA, ModelNode nodeB,
                                                 Map<ModelNode, Set<SimplePhysicalEntity>> nodeToEntity,
                                                 Model model, Integer graphLimit,
                                                 Map<ModelNode, Set<String>> blacklists) {

        Set<BioPAXElement> source = new HashSet<BioPAXElement>(nodeToEntity.get(nodeA));
        Set<BioPAXElement> target = new HashSet<BioPAXElement>(nodeToEntity.get(nodeB));
        Distance distanceResult = new Distance(nodeA, nodeB, 0, null);

        HashSet<String> blacklist = new HashSet<String>();
        Set<String> blacklistA = blacklists.get(nodeA);
        Set<String> blacklistB = blacklists.get(nodeB);

        if(blacklistA != null)
            blacklist.addAll(blacklistA);
        if(blacklistB != null)
            blacklist.addAll(blacklistB);

        for(int tmpLimit = 1; tmpLimit <= graphLimit; tmpLimit++) {
            UbiqueFilter ubiqueFilter = new UbiqueFilter(blacklist);
            Set<BioPAXElement> result
                    = QueryExecuter.runPathsFromTo(source, target, model, LimitType.NORMAL, tmpLimit, ubiqueFilter);

            if(!result.isEmpty()) {
                boolean isProperPath = false;
                Model distanceModel = BioPAXLevel.L3.getDefaultFactory().createModel();
                for (BioPAXElement bioPAXElement : result) {
                    distanceModel.add(bioPAXElement);
                    if(bioPAXElement instanceof Interaction)
                        isProperPath = true;
                }

                if(isProperPath) {
                    distanceResult.setDistance(tmpLimit);
                    distanceResult.setModel(distanceModel);
                    break;
                }
            }
        }

        return distanceResult;
    }

    public static boolean hasEntityFeatures(ModelNode modelNode, Set<EntityFeature> entityFeatures, boolean strict_mode) {
        SimplePhysicalEntity spe = BioPAXLevel.L3.getDefaultFactory().create(Protein.class, "dummyprotein");

        for (EntityFeature entityFeature : entityFeatures)
            spe.addFeature(entityFeature);

        boolean hasEF = hasEquivalentStates(modelNode, spe, strict_mode);

        for (EntityFeature entityFeature : entityFeatures)
            spe.removeFeature(entityFeature);

        return hasEF;
    }
}
