/*
 * Copyright (c) 2012 Memorial-Sloan Kettering Cancer Center.
 *
 * This file is part of BP_Prior.
 *
 * BP_Prior is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BP_Prior is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with BP_Prior.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mskcc.cbio.piclub.util;

import org.biopax.paxtools.io.SimpleIOHandler;
import org.biopax.paxtools.model.BioPAXFactory;
import org.biopax.paxtools.model.BioPAXLevel;
import org.biopax.paxtools.model.Model;
import org.biopax.paxtools.model.level3.*;
import org.mskcc.cbio.piclub.model.ModelNode;
import org.mskcc.cbio.piclub.model.Phosphorylation;

import org.junit.Test;

import java.io.InputStream;
import java.util.List;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertTrue;

public class PathwayCommons2UtilTest {
    private BioPAXFactory bioPAXFactory = BioPAXLevel.L3.getDefaultFactory();

    @Test
    public void hasEquivalentStatesStrictTest() {
        String geneName = "GENE";
        int phosphoLoc = 42, phosphoLoc2 = 144;
        Phosphorylation.Residue residue = Phosphorylation.Residue.SERINE,
                residue2 = Phosphorylation.Residue.THREONINE;
        String residueStr = residue.getAbbr(), residueStr2 = residue2.getAbbr();
        String phosphoName = geneName + "_p" + residueStr + phosphoLoc,
               state = "p" + residueStr + phosphoLoc,
               phosphoName2 = geneName + "_p" + residueStr2 + phosphoLoc2,
               state2 = "p" + residueStr2 + phosphoLoc2;


        ModelNode normalActiveNode = new ModelNode(geneName, null, geneName, ModelNode.Type.ACTIVATED);
        ModelNode phosphoActiveNode = new ModelNode(phosphoName, state, geneName, ModelNode.Type.ACTIVATED);
        ModelNode phosphoActiveNode2 = new ModelNode(phosphoName2, state2, geneName, ModelNode.Type.ACTIVATED);

        // First create a non-modified protein
        Model model = bioPAXFactory.createModel();
        RelationshipXref relationshipXref
                = bioPAXFactory.create(RelationshipXref.class, PathwayCommons2Util.convertToURI(geneName));
        ProteinReference proteinReference
                = bioPAXFactory.create(ProteinReference.class, "ref:" + geneName);
        Protein protein = bioPAXFactory.create(Protein.class, "protein:" + geneName);
        proteinReference.addXref(relationshipXref);
        protein.setEntityReference(proteinReference);
        model.add(protein);
        model.add(proteinReference);
        model.add(relationshipXref);

        ModificationFeature modificationFeature = bioPAXFactory.create(ModificationFeature.class, "mf:" + geneName);
        SequenceSite sequenceSite = bioPAXFactory.create(SequenceSite.class, "seqsite:" + geneName);
        sequenceSite.setSequencePosition(phosphoLoc);
        SequenceModificationVocabulary sequenceModificationVocabulary
                = bioPAXFactory.create(SequenceModificationVocabulary.class, "smvocab:" + geneName);
        String term = "O-phospho-L-" + residue.getName();
        sequenceModificationVocabulary.addTerm(term);
        modificationFeature.setModificationType(sequenceModificationVocabulary);
        modificationFeature.setFeatureLocation(sequenceSite);
        protein.addFeature(modificationFeature);
        model.add(sequenceSite);
        model.add(sequenceModificationVocabulary);
        model.add(modificationFeature);

        ModificationFeature modificationFeature1 = bioPAXFactory.create(ModificationFeature.class, "mf1:" + geneName);
        SequenceModificationVocabulary sequenceModificationVocabulary1
                = bioPAXFactory.create(SequenceModificationVocabulary.class, "smvocab1:" + geneName);
        String term1 = "residue modification, active";
        sequenceModificationVocabulary1.addTerm(term1);
        modificationFeature1.setModificationType(sequenceModificationVocabulary1);
        protein.addFeature(modificationFeature1);
        model.add(modificationFeature1);
        model.add(sequenceModificationVocabulary1);


        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein, false));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode2, protein, false));

        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein, true));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode2, protein, true));


        ModificationFeature modificationFeature2 = bioPAXFactory.create(ModificationFeature.class, "mf2:" + geneName);
        SequenceSite sequenceSite2 = bioPAXFactory.create(SequenceSite.class, "seqsite2:" + geneName);
        sequenceSite2.setSequencePosition(phosphoLoc2);
        SequenceModificationVocabulary sequenceModificationVocabulary2
                = bioPAXFactory.create(SequenceModificationVocabulary.class, "smvocab2:" + geneName);
        String term2 = "O-phospho-L-" + residue2.getName();
        sequenceModificationVocabulary2.addTerm(term2);
        modificationFeature2.setModificationType(sequenceModificationVocabulary2);
        modificationFeature2.setFeatureLocation(sequenceSite2);
        protein.addFeature(modificationFeature2);
        model.add(sequenceSite2);
        model.add(sequenceModificationVocabulary2);
        model.add(modificationFeature2);

        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein, false));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode2, protein, false));

        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein, true));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode2, protein, true));

        // Now, just the active tag
        protein.removeFeature(modificationFeature);
        protein.removeFeature(modificationFeature2);

        assertTrue(PathwayCommons2Util.hasEquivalentStates(normalActiveNode, protein, false));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(normalActiveNode, protein, true));
    }


    @Test
    public void hasEquivalentStatesTest() {
        String geneName = "GENE";
        int phosphoLoc = 42;
        Phosphorylation.Residue residue = Phosphorylation.Residue.SERINE;
        String residueStr = residue.getAbbr();
        String phosphoName = geneName + "_p" + residueStr + phosphoLoc;
        String state = "p" + residueStr + phosphoLoc;

        ModelNode concNode = new ModelNode(geneName, null, geneName, ModelNode.Type.CONCENTRATION);
        ModelNode phosphoActiveNode = new ModelNode(phosphoName, state, geneName, ModelNode.Type.ACTIVATED);
        ModelNode phosphoInactiveNode = new ModelNode(phosphoName, state, geneName, ModelNode.Type.INHIBITED);
        ModelNode activeNode = new ModelNode(geneName, null, geneName, ModelNode.Type.ACTIVATED);
        ModelNode inactiveNode = new ModelNode(geneName, null, geneName, ModelNode.Type.INHIBITED);

        // First create a non-modified protein
        Model model = bioPAXFactory.createModel();
        RelationshipXref relationshipXref
                = bioPAXFactory.create(RelationshipXref.class, PathwayCommons2Util.convertToURI(geneName));
        ProteinReference proteinReference
                = bioPAXFactory.create(ProteinReference.class, "ref:" + geneName);
        Protein protein = bioPAXFactory.create(Protein.class, "protein:" + geneName);
        proteinReference.addXref(relationshipXref);
        protein.setEntityReference(proteinReference);
        model.add(protein);
        model.add(proteinReference);
        model.add(relationshipXref);

        // This should be mapped to concentration nodes
        assertTrue(PathwayCommons2Util.hasEquivalentStates(concNode, protein));
        // and not to the others
        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoInactiveNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(activeNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(inactiveNode, protein));

        // Now, let's add a phosphorylation
        ModificationFeature modificationFeature = bioPAXFactory.create(ModificationFeature.class, "mf:" + geneName);
        SequenceSite sequenceSite = bioPAXFactory.create(SequenceSite.class, "seqsite:" + geneName);
        sequenceSite.setSequencePosition(phosphoLoc);
        SequenceModificationVocabulary sequenceModificationVocabulary
                = bioPAXFactory.create(SequenceModificationVocabulary.class, "smvocab:" + geneName);
        String term1 = "O-phospho-L-" + residue.getName();
        sequenceModificationVocabulary.addTerm(term1);
        modificationFeature.setModificationType(sequenceModificationVocabulary);
        modificationFeature.setFeatureLocation(sequenceSite);
        protein.addFeature(modificationFeature);
        model.add(sequenceSite);
        model.add(sequenceModificationVocabulary);
        model.add(modificationFeature);

        // Now it should match w/ concentration, phosphoActive and phosphoInactive ones
        assertTrue(PathwayCommons2Util.hasEquivalentStates(concNode, protein));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoInactiveNode, protein));
        // and not to the others
        assertFalse(PathwayCommons2Util.hasEquivalentStates(activeNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(inactiveNode, protein));

        // Now, let's add another modification, but this time let it be ambigous
        ModificationFeature modificationFeature2 = bioPAXFactory.create(ModificationFeature.class, "mf2:" + geneName);
        SequenceModificationVocabulary sequenceModificationVocabulary2
                = bioPAXFactory.create(SequenceModificationVocabulary.class, "smvocab2:" + geneName);
        String term2 = "residue modification, active";
        sequenceModificationVocabulary2.addTerm(term2);
        modificationFeature2.setModificationType(sequenceModificationVocabulary2);
        protein.addFeature(modificationFeature2);
        model.add(modificationFeature2);
        model.add(sequenceModificationVocabulary2);

        // Since we add the "active" annotation, now it should only match with conc, phosphoactive and active
        assertTrue(PathwayCommons2Util.hasEquivalentStates(concNode, protein));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(activeNode, protein));
        // and not to the others
        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoInactiveNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(inactiveNode, protein));

        // Now, let's make it inactive this time
        ModificationFeature modificationFeature3 = bioPAXFactory.create(ModificationFeature.class, "mf3:" + geneName);
        SequenceModificationVocabulary sequenceModificationVocabulary3
                = bioPAXFactory.create(SequenceModificationVocabulary.class, "smvocab3:" + geneName);
        String term3 = "residue modification, inactive";
        sequenceModificationVocabulary3.addTerm(term3);
        modificationFeature3.setModificationType(sequenceModificationVocabulary3);
        // Remove the active tag
        protein.removeFeature(modificationFeature2);
        // And add the inactive
        protein.addFeature(modificationFeature3);
        model.add(modificationFeature3);
        model.add(sequenceModificationVocabulary3);

        // Should now match w/ phosphoInactive and inactive
        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoInactiveNode, protein));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(inactiveNode, protein));
        // and not to the others
        assertFalse(PathwayCommons2Util.hasEquivalentStates(concNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(activeNode, protein));

        // Finally, let's change the phosphorylation site
        sequenceSite.setSequencePosition(phosphoLoc + 100);

        // Should now match only w/ inactive ones
        assertTrue(PathwayCommons2Util.hasEquivalentStates(inactiveNode, protein));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(phosphoInactiveNode, protein));
        // and not to the others
        assertFalse(PathwayCommons2Util.hasEquivalentStates(concNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(phosphoActiveNode, protein));
        assertFalse(PathwayCommons2Util.hasEquivalentStates(activeNode, protein));

        // OK, there are a couple of other situations, but for now that's it!
        // TODO: Add more states (how about fragmentation, binding?..)
    }

    // The order the features should not effect the node -> entity mapping
    @Test
    public void featureOrderEffectTest() throws Exception {
        InputStream fileStream = this.getClass().getClassLoader().getResourceAsStream("singleSPE.owl");
        SimpleIOHandler ioHandler = new SimpleIOHandler();
        Model model = ioHandler.convertFromOWL(fileStream);

        fileStream = this.getClass().getClassLoader().getResourceAsStream("singleSPE_nodes.tsv");
        List<ModelNode> modelNodes = ModelNodeUtil.parseGeneFile(fileStream);
        assertEquals(2, modelNodes.size());

        Protein protein = (Protein) model.getByID("HTTP://WWW.REACTOME.ORG/BIOPAX/48887#PROTEIN1644_1_9606");
        assertTrue(PathwayCommons2Util.hasEquivalentStates(modelNodes.get(0), protein));
        assertTrue(PathwayCommons2Util.hasEquivalentStates(modelNodes.get(1), protein));
    }

    @Test
    public void convertToURITest() {
        Model model = bioPAXFactory.createModel();
        model.add(bioPAXFactory.create(RelationshipXref.class, "urn:biopax:RelationshipXref:HGNC_HGNC%3A11998"));
        assertTrue(model.getByID(PathwayCommons2Util.convertToURI("TP53")) != null);
    }
}
